﻿package meyermike.core
{
    import flash.display.Bitmap;
    import flash.display.DisplayObject;
    import flash.display.DisplayObjectContainer;
    import flash.display.Shape;
    import flash.display.Sprite;
    import flash.display.Stage;
    import flash.display.StageScaleMode;
    import flash.events.Event;
    import flash.events.KeyboardEvent;
    import flash.geom.Point;
    import flash.geom.Rectangle;
    import meyermike.core.GameInterface;
    import meyermike.core.GameObject;
    import meyermike.debug.DebugWidget;
    import meyermike.Global;
    import meyermike.InputUtil;
    import meyermike.util.MathUtil;
    import meyermike.util.PhysicsUtil;
    import meyermike.screen.Pause;
    import meyermike.core.GameState;
    import meyermike.Util;
    import meyermike.util.CheatUtil;
    import meyermike.util.music.MusicUtil;
    import meyermike.util.SceneLayers;
    import meyermike.util.TimeUtil;
    import meyermike.widget.MusicCredits;
    import meyermike.widget.TextField;

    /**
     * Base class that you will be extending to create your own games.
     * @author Mike Meyer
     */
    public class BaseGame extends Sprite implements GameInterface
    {
        protected var mScript:GameState;
        private var mNextScript:GameState;
        private var mIsPaused:Boolean;
        private static const CAMERA_TIGHTNESS:Number = 0.125;

        /// widget for music credits
        protected var mMusicCredits:MusicCredits;

        //private var _cameraFollowTarget:GameObject;
        //public function get CameraFollowTarget():GameObject { return _cameraFollowTarget; }
        //public function set CameraFollowTarget(go:GameObject):void { _cameraFollowTarget = go; }

        private var _cameraPosition:Point = new Point(0,0);
        public function get CameraPosition():Point { return _cameraPosition; }
        public function set CameraPosition(pos:Point):void { _cameraPosition.x = pos.x; _cameraPosition.y = -pos.y;  }

        /// see how much the game is zoomed in
        public function get Zoom():Number { return mGameHolder.scaleX; }

        /// set how much the game is zoomed in
        public function set Zoom(zoom:Number):void { mGameHolder.scaleX = zoom; mGameHolder.scaleY = zoom; }

        private var mPauseScreen:Sprite;
        private var mGameHolder:DisplayObjectContainer;

        private var _cheatUtil:CheatUtil;
        public function get Cheats():CheatUtil { return _cheatUtil; }

        private var _musicUtil:MusicUtil;
        public function get Music():MusicUtil { return _musicUtil; }

        private var _layers:SceneLayers;
        public function get Layers():SceneLayers { return _layers; }

        private var _time:TimeUtil;
        public function get Time():TimeUtil { return _time; }

        private var _physics:PhysicsUtil;
        public function get Physics():PhysicsUtil { return _physics; }

        private var _debugger:DebugWidget;
        public function get Debug():DebugWidget { return _debugger; }

        private var _cameraShake:Number = 0;
        public function get CameraShake():Number { return _cameraShake; }
        public function set CameraShake(value:Number):void { _cameraShake = value;  }

        // shake looks much better if the shaking tends to
        // go back and forth instead of actually random
        private var _lastShakeDir:Number = 0;
        private static const MIN_SHAKE_ROT:Number = 0.67 * Math.PI;
        private static const MAX_SHAKE_ROT:Number = 1.33 * Math.PI;

        /**
         * Constructor
         */
        public function BaseGame():void
        {
            super();

            if (stage)
            {
                Init();
            }
            else
            {
                addEventListener(Event.ADDED_TO_STAGE, _Init);
            }
        }

        /**
         * Automagically called by the constructor for the initial init,
         * but if you shut it down and want to restart, you'll need to call it yourself.
         *
         * When inheriting from BaseGame, it is recommended that you instead override
         * PreloadComplete
         */
        public function Init():void
        {
            // hacks to make things work whatever the window size is
            stage.scaleMode = StageScaleMode.NO_SCALE;
            Global.StageWidth = 480;
            Global.StageHeight = 320;

            // TODO: verify that it goes through the preloader
            // make BaseGame do the preloading.  Then once it
            // is done, call Init
            PreloadComplete();
        }

        /**
         * light wrapper for making Init usable as an event callback
         */
        public function _Init(e:Event):void
        {
            removeEventListener(Event.ADDED_TO_STAGE, _Init);
            Init();
        }

        /**
         * Preloading has finished! Start the game!
         */
        protected function PreloadComplete():void
        {
            _cheatUtil = new CheatUtil();
            Cheats.Init(stage);

            _musicUtil = new MusicUtil();
            Music.Init();

            _time = new TimeUtil();
            Time.Init();

            _physics = new PhysicsUtil();

            //CameraFollowTarget = null;

            mIsPaused = false;

            InputUtil.Init(stage);

            addEventListener(Event.ENTER_FRAME, HandleEnterFrame);

            SetupScene();

            mNextScript = null;

            stage.addEventListener(Event.DEACTIVATE, _Pause);
            stage.addEventListener(Event.ACTIVATE, _Unpause);
        }

        /**
         * add the persistent display objects that stick around even between levels
         */
        private function SetupScene():void
        {
            // holder for the game itself (gets moved to scroll the camera)
            mGameHolder = new Sprite();
            this.addChild(mGameHolder);

            _layers = new SceneLayers();
            Layers.Init(mGameHolder);

            _debugger = new DebugWidget();
            this.addChild(_debugger);
            _debugger.Init();

            // display for the music credits
            mMusicCredits = new MusicCredits(Music);
            mMusicCredits.Init();
            mMusicCredits.y = Global.StageHeight - mMusicCredits.height;
            this.addChild(mMusicCredits);

            CreatePauseIndicator();
        }

        /**
         * create the pause screen
         */
        protected function CreatePauseIndicator():void
        {
            mPauseScreen = new meyermike.screen.Pause();
            this.addChild(mPauseScreen);
        }

        protected function HandleEnterFrame(e:Event = null):void
        {
            Time.TimeDelta = (1.0 / stage.frameRate);

            ScriptChange();

            if (!mIsPaused)
            {
                Update();
                Integrate();
                PreDraw();
            }
        }

        protected function Update():void
        {
            mScript.Update();
            Physics.Update(Layers, Time.TimeDelta);
            Layers.Update(Time.TimeDelta);
            InputUtil.Update();
            mMusicCredits.Update(Time.TimeDelta);

            // camera shake/positioning
            // technically, for a few lines, _lastShakeDir is representing the future shake dir
            _lastShakeDir += MathUtil.Lerp( MIN_SHAKE_ROT, MAX_SHAKE_ROT, Math.random() );
            _lastShakeDir = MathUtil.Wrap( _lastShakeDir, 0, 2.0 * Math.PI);
            mGameHolder.x = -CameraPosition.x + (CameraShake * Math.cos(_lastShakeDir));
            mGameHolder.y = -CameraPosition.y + (CameraShake * Math.sin(_lastShakeDir));
        }

        protected function Integrate():void
        {
            Layers.Integrate(Time.TimeDelta);
        }
        protected function PreDraw():void
        {
            Layers.PreDraw();
        }

        /// Called on each Beat of the game.
        protected function Beat(e:Event = null):void
        {
            var beat:Number = mScript.GetCurBeat();

            Layers.Beat(beat);
        }

        protected function OffBeat(e:Event = null):void
        {
            var beat:Number = mScript.GetCurBeat();

            Layers.OffBeat(beat);
        }
        protected function QuarterBeat(e:Event = null):void
        {
            var beat:Number = mScript.GetCurQuarterBeat();

            Layers.QuarterBeat(beat);
        }

        /**
         * freeze time and show the pause screen
         */
        public function Pause():void
        {
            mIsPaused = true;
            mPauseScreen.visible = true;
            Music.Pause();
        }

        /**
         * light wrapper for using Pause as an event callback
         */
        private function _Pause(e:Event):void
        {
            Pause();
        }

        /**
         * resume after pausing
         */
        public function Unpause():void
        {
            mIsPaused = false;
            mPauseScreen.visible = false;
            Music.Resume();
        }

        /**
         * light wrapper for using Pause as an event callback
         */
        private function _Unpause(e:Event):void
        {
            Unpause();
        }

        /**
         * if a script change was queued up, perform the actual switch now
         */
        private function ScriptChange():void
        {
            if (mNextScript != null)
            {
                try
                {
                    EndScript(mScript);
                    BeginScript(mNextScript);

                    mScript = mNextScript;
                    mNextScript = null;
                }
                catch (e:Error)
                {
                    // don't try to do that again
                    mNextScript = null;
                    // rethrow
                    throw e;
                }
            }
        }

        /**
         * shut it down
         */
        private function EndScript(script:GameState):void
        {
            if (script != null)
            {
                script.removeEventListener(GameState.EVENT_BEAT, Beat);
                script.removeEventListener(GameState.EVENT_OFFBEAT, OffBeat);
                script.removeEventListener(GameState.EVENT_QUARTERBEAT, QuarterBeat);
                script.End();
            }
        }

        /**
         *
         */
        private function BeginScript(script:GameState):void
        {
            if (script != null)
            {
                script.addEventListener(GameState.EVENT_BEAT, Beat);
                script.addEventListener(GameState.EVENT_OFFBEAT, OffBeat);
                script.addEventListener(GameState.EVENT_QUARTERBEAT, QuarterBeat);
                script.Begin(this);
            }
        }

        /**
         * set a script to switch to once the current beat/update is finished
         */
        public function ChangeGameState(nextScript:GameState):void
        {
            mNextScript = nextScript;
        }

        /**
         * go back to the starting position
         */
        public function ResetScrolling():void
        {
            mGameHolder.x = 0;
            mGameHolder.y = 0;
        }

        /**
         * Clean up
         */
        public function Shutdown():void
        {
            //CameraFollowTarget = null;
            mScript = null;
            mNextScript = null;

            stage.removeEventListener(Event.DEACTIVATE, _Pause);
            stage.removeEventListener(Event.ACTIVATE, _Unpause);

            if (null != mGameHolder)
            {
                this.removeChild(mGameHolder);
                mGameHolder = null;
            }

            if (null != mPauseScreen)
            {
                this.removeChild(mPauseScreen);
                mPauseScreen = null;
            }

            if (null != this.Cheats)
            {
                this.Cheats.Shutdown();
                _cheatUtil = null;
            }

            if (null != this.Music)
            {
                this.Music.Shutdown();
                _musicUtil = null;
            }

            if (null != this.Layers)
            {
                this.Layers.Shutdown();
                _layers = null;
            }

            if (null != Time)
            {
                Time.Shutdown();
                _time = null;
            }
        }
    }
}
